Index: sys/mips/include/pte.h
===================================================================
--- sys/mips/include/pte.h	(revision 209521)
+++ sys/mips/include/pte.h	(working copy)
@@ -73,7 +73,8 @@
  * Note that in FreeBSD, we map 2 TLB pages is equal to 1 VM page.
  */
 #define	TLBHI_ASID_MASK		(0xff)
-#define	TLBHI_ENTRY(va, asid)	(((va) & ~PAGE_MASK) | ((asid) & TLBHI_ASID_MASK))
+#define	TLBHI_PAGE_MASK		(2 * PAGE_SIZE - 1)
+#define	TLBHI_ENTRY(va, asid)	(((va) & ~TLBHI_PAGE_MASK) | ((asid) & TLBHI_ASID_MASK))
 
 #ifndef _LOCORE
 typedef	uint32_t pt_entry_t;
Index: sys/mips/mips/tlb.c
===================================================================
--- sys/mips/mips/tlb.c	(revision 209521)
+++ sys/mips/mips/tlb.c	(working copy)
@@ -217,30 +217,41 @@
 void
 tlb_update(struct pmap *pmap, vm_offset_t va, pt_entry_t pte)
 {
-	register_t mask, asid;
+	pt_entry_t other;
+	register_t mask, asid, hi;
 	register_t s;
-	int i;
+	int i, even;
 
-	va &= ~PAGE_MASK;
+	s = intr_disable();
+	even = (va & PAGE_SIZE) == 0;
 	pte &= ~TLBLO_SWBITS_MASK;
-
-	s = intr_disable();
+	hi = TLBHI_ENTRY(va, pmap_asid(pmap));
 	mask = mips_rd_pagemask();
 	asid = mips_rd_entryhi() & TLBHI_ASID_MASK;
-
-	mips_wr_pagemask(0);
-	mips_wr_entryhi(TLBHI_ENTRY(va, pmap_asid(pmap)));
+	mips_wr_entryhi(hi);
 	tlb_probe();
 	i = mips_rd_index();
 	if (i >= 0) {
 		tlb_read();
 
-		if ((va & PAGE_SIZE) == 0) {
+		if (even) {
 			mips_wr_entrylo0(pte);
 		} else {
 			mips_wr_entrylo1(pte);
 		}
 		tlb_write_indexed();
+	} else {
+		other = pte & PTE_G;
+		mips_wr_pagemask(0);
+		mips_wr_entryhi(hi);
+		if (even) {
+			mips_wr_entrylo0(pte);
+			mips_wr_entrylo1(other);
+		} else {
+			mips_wr_entrylo0(other);
+			mips_wr_entrylo1(pte);
+		}
+		tlb_write_random();
 	}
 
 	mips_wr_entryhi(asid);